// To compile, perform:
// gcc -std=c99 randomQuickSort_MersenneTwister_and_heap.c -o randomQuickSort_MersenneTwister_and_heap
// Run:
// ./randomQuickSort_MersenneTwister_and_heap

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>
#include "MT.h"     // Use the Mersenne Twister.

int partition(int* target, int left, int right) {
    // Generate a random number, genrand_int32(), using the the Mersenne Twister.
    int random = left + genrand_int32() % (right - left + 1);
    // Exchange target[right] and target[random].
    int tmp = target[right]; target[right] = target[random]; target[random] = tmp;
    int pivot = target[right];
    int i = left-1; // i scans the array from the left.
    int j = right;  // j scans the array from the right.
    for (;;) {
        // Move from the left until hitting a value no less than the pivot.
        for(i++; target[i] < pivot; i++){}
        // Move from the right until hitting a value no greater than the pivot.
        for(j--; pivot < target[j] && i < j; j--){}
        if (i >= j)  break;
        // Exchange target[i] and target[j].
        tmp = target[i];  target[i] = target[j];  target[j] = tmp;
    }
    // Exchange target[i] and target[right].
    tmp = target[i];  target[i] = target[right];  target[right] = tmp;
    return i;
}

void randomQuickSort1(int* target, int left, int right ) {
    if (left < right) {
        int i = partition(target, left, right); // i: Position of the pivot.
        randomQuickSort1(target, left, i - 1);
        randomQuickSort1(target, i + 1, right);
    }
}

void randomQuickSort2(int* target, int aLeft, int right) {
    int left = aLeft;
    while (left < right) {
        int i = partition(target, left, right);
        randomQuickSort2(target, left, i-1);
        left = i+1;
    }
}

void randomQuickSort3(int* target, int aLeft, int aRight) {
    int left = aLeft; int right = aRight;
    while (left < right) {
        int i = partition(target, left, right);
        if( i - left <= right - i ){ // The left interval is shorter.
            randomQuickSort3(target, left, i-1);
            left=i+1;
        }else{                       // The right interval is shorter.
            randomQuickSort3(target, i+1, right);
            right=i-1;
        }
    }
}

int main(int argc, char *argv[])
{
    // Let us use the heap for the target array.
    int inputSize = 1000000;   // Specify the size of the input array, target.
    int *target  = (int *)malloc(sizeof(int) * inputSize);  // Use the heap.
    if(target == NULL)      // Failure to allocate the main memory for target.
        exit(EXIT_FAILURE); // Try instead, say, inputSize = 1000000000000;
    for(int i=0; i<inputSize; i++)
        target[i] = genrand_int32() % 524287;   // Fill the array with random numbers from 0 to 2^19 - 1
    
    struct timeval s, e;    // timeval has tv_sec (second) and tv_usec (micro-seconds).
    gettimeofday(&s, NULL); // Get the time to begin.
    randomQuickSort3(target, 0, inputSize-1);   // Sort the target array.
    gettimeofday(&e, NULL); // Get the time to end and calculate the computation time.
    float comp_time = (e.tv_sec - s.tv_sec) + (e.tv_usec - s.tv_usec)*1.0E-6;
    
    for(int i=0; i < inputSize-1; i++){ // Check if the target is sorted correctly.
        if( target[i] > target[i+1]){   // The two neighboring numers are NOT sorted.
            fprintf(stderr, "Failure to sort the target array.\n");
            free(target); exit(EXIT_FAILURE); }
    }
    printf("The input of %i in size was sorted successfully in %f sec.\n", inputSize, comp_time);
    
    free(target);
    return EXIT_SUCCESS;
}
